#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : counts.py
# Contents: count patterns by size and support in independent data
# Author  : Christian Borgelt
# History : 2013.05.09 file created from detect/counts.py
#           2013.12.06 filling holes replaced by detection border
#           2014.05.27 modules placed in sibling directories
#-----------------------------------------------------------------------
from sys    import argv, path
from os.path import join
from math   import floor
from time   import strftime
from random import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par import run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts   import genpsts
try:
    from coco    import coconad
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

def counts (ncnt=100, period=3, rate=20, width=0.003,
            target='c', runs=1000, verbose=100, srng=1):
    if srng == 0: srng = 1      # init. seed and get data file name
    name = 'n%d_p%g_r%g_w%02d_%c' \
         % (ncnt,period,rate,int(floor(width*10000+0.5)),target)
    psp  = dict()               # create a pattern spectrum
    for i in range(runs):       # repeat the experiments multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S counts ')+name+' '+str(i))
        seed(srng)              # seed random number generator
        psts = list(zip(range(1,ncnt+1), genpsts(ncnt, period, rate)))
        cps  = coconad(psts, target=target, width=width,
                       supp=2, zmin=2, report='#')
        for s in cps:           # generate and mine spike trains
            if s in psp: psp[s] += cps[s]
            else:        psp[s]  = cps[s]
        # slower alternative to a pattern spectrum
        # pats = coconad(psts, target=target, width=width,
        #                supp=2, zmin=2)
        # for p,r in pats:        # generate and mine spike trains
        #     s = (len(p),r)      # and traverse the found patterns
        #     if s in psp: psp[s] += 1
        #     else:        psp[s]  = 1
        srng += 2               # update the pattern counters
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S counts ')
                      +name+' '+str(runs)+' done.')
    n = 1.0/runs if runs > 0 else 1.0
    with open(join('data','pc_'+name+'.txt'), 'w') as out:
        for x in sorted([(s[0], s[1], psp[s]*n) for s in psp]):
            out.write('%g %g %g\n' % x)
    return psp                  # write and return pattern counters

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'target', [  'c' ] ),
                ( 'width',  [ 0.0060, 0.0058, 0.0056, 0.0054, 0.0052,
                              0.0050, 0.0048, 0.0046, 0.0044, 0.0042,
                              0.0040, 0.0038, 0.0036, 0.0034, 0.0032,
                              0.0030, 0.0028, 0.0026, 0.0024, 0.0022,
                              0.0020, 0.0018, 0.0016, 0.0014, 0.0012,
                              0.0010 ] ),
                ( 'runs',   [ 1000 ] ) ] ]

    run(counts, grids, int(argv[1])-1 if len(argv) > 1 else -1)
