#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : detect.py
# Contents: determine best window width for CoCoNAD algorithm
# Author  : Christian Borgelt
# History : 2013.05.09 file created from detect/detect.py
#           2014.05.27 modules placed in sibling directories
#           2015.08.12 adapted to modified pattern set reduction
#           2015.08.18 changed to function patred() from pycoco library
#-----------------------------------------------------------------------
from sys     import argv, path
from os.path import join
from math    import floor
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import genpsts
from pcseval import getpcsfn
try:
    from coco    import coconad, patred
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    FIMDIR   = join('..', 'fim')
    if FIMDIR   not in path: path.insert(0, FIMDIR)
    from patred  import patred
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,         # neuron    parameters
            size=5, coin=5, jitter=0.003,        # assembly  parameters
            widths=[0.003], target='c',          # detection parameters
            runs=1000, verbose=100, srng=1):     # other     parameters
    seed(srng +1023*size)       # create the assembly to inject and
    asm  = set(range(1,size+1)) # generate parallel spike trains
    psts = genpsts(ncnt,period,rate, jitter=jitter, asms=[[asm,-coin]])
    psts = list(zip(range(1,ncnt+1), psts))
    pre  = 'n%d_p%g_r%g' % (ncnt,period,rate)
    sjit = '_j%g'   % (jitter*1000)
    post = '_%c'    % target    # form the file name (parts)
    ext  = '_%d_%d' % (size,coin)
    dect = []                   # init. the result array
    for w in widths:            # traverse the window widths
        swd  = '_w%02d' % int(floor(w*10000+0.5))
        name = pre+sjit+swd+post+ext
        if verbose > 0:         # print a log message
            print(strftime('%Y.%m.%d %H:%M:%S detect ')+name)
        psp = set()             # initialize a pattern spectrum
        with open(join('data','pc_'+pre+swd+post+'.txt'), 'r') as inp:
            for line in inp:    # collect the pattern signatures
                z,c,n = line.split(); z,c = int(z),int(c)
                if z > 1 and c > 1: psp.add((z,c))
        border = list(range(ncnt+1)); m = -1
        for z in reversed(border):  # determine the detection border
            m = max([c for a,c in psp if a == z]+[m])
            border[z] = m+1     # (support thresholds per size)
        border[0:2] = [oo,oo]   # entirely rule out sizes 0 and 1
        for z in range(2,len(border)):
            for c in range(2,border[z]):
                psp.add((z,c))  # fill holes in the pattern spectrum
        pats = coconad(psts, target=target, width=w, supp=2, zmin=2)
        fltd = []               # mine the created spike trains and
        upcs = dict()           # initialize the pattern spectra
        fpcs = dict()           # (unfiltered, filtered, reduced)
        rpcs = dict()
        for p,c in pats:        # and traverse the found patterns
            z = len(p); s = (z,c)      # create the pattern signature
            if z < 2 or c < 2: continue
            if s in upcs: upcs[s] += 1 # count unfiltered patterns
            else:         upcs[s]  = 1 # by their signatures
            if c < border[z]: continue # collect non-excluded patterns
            fltd.append((frozenset(p),c))
            if s in fpcs: fpcs[s] += 1 # count filtered patterns
            else:         fpcs[s]  = 1 # by their signatures
        pats = patred(fltd, 'S', border, True)
        for p,c in pats:        # traverse the reduced patterns and
            s = (len(p),c)      # count them by their signatures
            if s in rpcs: rpcs[s] += 1 # count reduced patterns
            else:         rpcs[s]  = 1 # by their signatures
        with open(join('data','su_'+name+'.txt'), 'w') as out:
            for x in sorted([(s[0], s[1], upcs[s]) for s in upcs]):
                out.write('%g %g %g\n' % x)
        with open(join('data','sf_'+name+'.txt'), 'w') as out:
            for x in sorted([(s[0], s[1], fpcs[s]) for s in fpcs]):
                out.write('%g %g %g\n' % x)
        with open(join('data','sr_'+name+'.txt'), 'w') as out:
            for x in sorted([(s[0], s[1], rpcs[s]) for s in rpcs]):
                out.write('%g %g %g\n' % x)
        x = [getpcsfn(i)(upcs, psp) for i in range(getpcsfn(-1))]
        dect.append((w, x))     # collect the evaluation results
    name = pre+sjit+post+ext    # form the output file name
    if verbose > 0:             # print a completion message
        print(strftime('%Y.%m.%d %H:%M:%S detect ')+name+' done.')
    with open(join('data','wd_'+name+'.txt'), 'w') as out:
        for w,r in dect:        # write the collected results
            out.write('%g' % w) # to an output file
            for e in r: out.write(' %g' % e)
            out.write('\n')     # (one width and results per line)
    return dect                 # return the results as a list

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( {'size': 5, 'coin': 5},
                  {'size': 6, 'coin': 6},
                  {'size': 7, 'coin': 7},
                  {'size': 8, 'coin': 8},
                  {'size': 9, 'coin': 9},
                  {'size':10, 'coin':10},
                  {'size':11, 'coin':11},
                  {'size':12, 'coin':12},
                  {'size': 5, 'coin': 9},
                  {'size': 9, 'coin': 5},
                  {'size': 6, 'coin':10},
                  {'size':10, 'coin': 6} ),
                ( 'jitter', [ 0.003] ),
                ( 'target', [  'c' ] ),
                ( 'widths', [ [ 0.0060, 0.0058, 0.0056, 0.0054, 0.0052,
                                0.0050, 0.0048, 0.0046, 0.0044, 0.0042,
                                0.0040, 0.0038, 0.0036, 0.0034, 0.0032,
                                0.0030, 0.0028, 0.0026, 0.0024, 0.0022,
                                0.0020, 0.0018, 0.0016, 0.0014, 0.0012,
                                0.0010 ] ] ),
                ( 'srng',   [    1 ] ) ] ]

    run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
